"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.APIClient = exports.APIError = void 0;
const tslib_1 = require("tslib");
const errors_1 = require("@oclif/errors");
const cli_ux_1 = require("cli-ux");
const lodash_1 = require("lodash");
const url = tslib_1.__importStar(require("url"));
const deps_1 = tslib_1.__importDefault(require("./deps"));
const login_1 = require("./login");
const request_id_1 = require("./request-id");
const vars_1 = require("./vars");
const debugFn = require("debug"); // eslint-disable-line quotes
const session_1 = require("./session");
const utils_1 = require("./utils");
const debug = debugFn(`api-client`);
class APIError extends errors_1.CLIError {
    constructor(httpError) {
        super(httpError);
        Object.defineProperty(this, "http", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "body", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        if (!httpError)
            throw new Error(`invalid error`);
        const options = httpError.body;
        if (!options || !options.message)
            throw httpError;
        const info = [];
        if (options.id)
            info.push(`Error ID: ${options.id}`);
        if (options.url)
            info.push(`See ${options.url} for more information.`);
        if (info.length)
            super([options.message, ``].concat(info).join(`\n`));
        else
            super(options.message);
        this.http = httpError;
        this.body = options;
    }
}
exports.APIError = APIError;
class APIClient {
    constructor(config, options = {}) {
        Object.defineProperty(this, "config", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: config
        });
        Object.defineProperty(this, "options", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: options
        });
        Object.defineProperty(this, "authPromise", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "http", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_login", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new login_1.Login(this.config, this)
        });
        Object.defineProperty(this, "_auth", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.config = config;
        if (options.required === undefined)
            options.required = true;
        options.preauth = options.preauth !== false;
        this.options = options;
        const apiUrl = new url.URL(vars_1.vars.apiUrl);
        const envHeaders = JSON.parse(process.env.RELAY_HEADERS || `{}`);
        const self = this;
        const opts = {
            host: apiUrl.hostname,
            port: apiUrl.port,
            protocol: apiUrl.protocol,
            headers: {
                accpet: `application/json`,
                'user-agent': this.config.userAgent,
                ...envHeaders,
            },
        };
        this.http = class APIHTTPClient extends deps_1.default.HTTP.HTTP.create(opts) {
            static trackRequestIds(response) {
                const responseRequestIdHeader = response.headers[request_id_1.requestIdHeader];
                if (responseRequestIdHeader) {
                    const requestIds = Array.isArray(responseRequestIdHeader) ? responseRequestIdHeader : responseRequestIdHeader.split(`,`);
                    request_id_1.RequestId.track(...requestIds);
                }
            }
            static async request(url, opts = {}, retries = 3) {
                opts.headers = opts.headers || {};
                opts.headers[request_id_1.requestIdHeader] = request_id_1.RequestId.create() && request_id_1.RequestId.headerValue;
                if (!Object.keys(opts.headers).find(h => h.toLowerCase() === `authorization`)) {
                    opts.headers.authorization = `Bearer ${self.auth}`;
                }
                retries--;
                try {
                    const response = await super.request(url, opts);
                    this.trackRequestIds(response);
                    return response;
                }
                catch (err) {
                    if (!(err instanceof deps_1.default.HTTP.HTTPError))
                        throw err;
                    if (retries > 0) {
                        if (opts.retryAuth !== false && err.http.statusCode === 401) {
                            debug(`Token expired`);
                            if (process.env.RELAY_API_KEY) {
                                throw new Error(`The token provided to RELAY_API_KEY is invalid. Please double-heck that you have the correct token, or run 'relay login' without RELAY_API_KEY set.`);
                            }
                            const tokens = session_1.getToken();
                            if (tokens === null || tokens === void 0 ? void 0 : tokens.refresh_token) {
                                debug(`attempting refresh`);
                                if (!self.authPromise)
                                    self.authPromise = self.refresh();
                                await self.authPromise;
                                debug(`refresh successful`);
                                opts.headers.authorization = `Bearer ${self.auth}`;
                                debug(`re-requesting original`);
                                return this.request(url, opts, retries);
                            }
                            else {
                                cli_ux_1.cli.log(`Not logged in`);
                                debug(`attempting login`);
                                if (!self.authPromise)
                                    self.authPromise = self.login();
                                await self.authPromise;
                                opts.headers.authorization = `Bearer ${self.auth}`;
                                debug(`re-requesting original`);
                                return this.request(url, opts, retries);
                            }
                        }
                    }
                    throw new APIError(err);
                }
            }
        };
    }
    get auth() {
        if (!this._auth) {
            if (process.env.RELAY_API_TOKEN && !process.env.RELAY_API_KEY)
                deps_1.default.cli.warn(`RELAY_API_TOKEN is set but you probably meant RELAY_API_KEY`);
            this._auth = process.env.RELAY_API_KEY;
            if (!this._auth) {
                const tokens = session_1.getToken();
                this._auth = tokens === null || tokens === void 0 ? void 0 : tokens.access_token;
            }
        }
        return this._auth;
    }
    set auth(token) {
        delete this.authPromise;
        this._auth = token;
    }
    refresh() {
        this._auth = undefined;
        return this._login.refresh();
    }
    get(url, options = {}) {
        return this.http.get(url, options);
    }
    post(url, options = {}) {
        return this.http.post(url, options);
    }
    put(url, options = {}) {
        return this.http.put(url, options);
    }
    delete(url, options = {}) {
        return this.http.delete(url, options);
    }
    request(url, options = {}) {
        return this.http.request(url, options);
    }
    login() {
        return this._login.login();
    }
    async logout() {
        try {
            await this._login.logout();
        }
        catch (err) {
            errors_1.warn(err);
        }
    }
    get defaults() {
        return this.http.defaults;
    }
    async whoami() {
        // TODO use parameterized subscriberId
        const subscriber = session_1.getDefaultSubscriber();
        const { body: user } = await this.get(`${vars_1.vars.authUrl}/oauth2/validate`);
        return {
            Name: `${user.given_name} ${user.family_name}`,
            Email: `${user.email}`,
            [`User ID`]: `${user.userid}`,
            [`Default Subscriber`]: subscriber.id,
        };
    }
    async devices(subscriberId) {
        const { body: { devices } } = await this.get(`/ibot/subscriber/${subscriberId}/device_ids`);
        return devices;
    }
    async groups(subscriberId) {
        let { body } = await this.get(`/ibot/groups?subscriber_id=${subscriberId}`);
        body = lodash_1.isString(body) ? JSON.parse(body) : body;
        return lodash_1.filter(body, { owner: subscriberId });
    }
    async workflows(subscriberId) {
        const { body: { results } } = await this.get(`/ibot/workflow?subscriber_id=${subscriberId}`);
        return lodash_1.map(results, row => {
            if (lodash_1.isString(row.config)) {
                row.config = JSON.parse(row.config);
            }
            return row;
        });
    }
    async workflow(subscriberId, id) {
        const workflows = await this.workflows(subscriberId);
        return lodash_1.find(workflows, ({ workflow_id }) => lodash_1.includes(workflow_id, id));
    }
    async saveWorkflow(workflow) {
        // TODO use parameterized subscriberId
        const subscriberId = session_1.getDefaultSubscriberId();
        const { body } = await this.post(`/ibot/workflow?subscriber_id=${subscriberId}`, {
            body: workflow,
        });
        return lodash_1.map(body, row => {
            if (lodash_1.isString(row.config)) {
                row.config = JSON.parse(row.config);
            }
            return row;
        });
    }
    async removeWorkflow(id) {
        // TODO use parameterized subscriberId
        const subscriberId = session_1.getDefaultSubscriberId();
        await this.delete(`/ibot/workflow/${id}?subscriber_id=${subscriberId}`);
        return true;
    }
    async triggerWorkflow(subscriberId, workflowId, args) {
        const tokenAccount = session_1.getToken();
        const uri = `/ibot/workflow/${workflowId}?subscriber_id=${subscriberId}&user_id=${tokenAccount === null || tokenAccount === void 0 ? void 0 : tokenAccount.uuid}`;
        const body = {
            action: `invoke`,
            action_args: args,
        };
        debug(`triggerWorkflow`, { uri, body });
        await this.post(uri, { body });
    }
    clear() {
        session_1.clearConfig();
        session_1.clearSubscribers();
    }
    session() {
        return session_1.getSession();
    }
    async subscribers() {
        const { body: accounts } = await this.request(`${vars_1.vars.stratusUrl}/v3/subscribers;view=dash_overview`);
        return lodash_1.map(accounts, account => ({
            id: utils_1.getOrThrow(account, [`account`, `subscriber_id`]),
            email: utils_1.getOrThrow(account, [`account`, `owner_email`]),
            name: utils_1.getOrThrow(account, [`account`, `account_name`]),
        }));
    }
}
exports.APIClient = APIClient;
