"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Login = void 0;
const tslib_1 = require("tslib");
const cli_ux_1 = tslib_1.__importDefault(require("cli-ux"));
const http_call_1 = tslib_1.__importDefault(require("http-call"));
const form_urlencoded_1 = tslib_1.__importDefault(require("form-urlencoded"));
const querystring_1 = tslib_1.__importDefault(require("querystring"));
const open_1 = tslib_1.__importDefault(require("open"));
const api_client_1 = require("./api-client");
const vars_1 = require("./vars");
const http_1 = tslib_1.__importDefault(require("http"));
const url_1 = require("url");
const lodash_1 = require("lodash");
const debug_1 = tslib_1.__importDefault(require("debug"));
const crypto = require("crypto"); // eslint-disable-line quotes
const session_1 = require("./session");
const utils_1 = require("./utils");
const debug = debug_1.default(`login`);
class Login {
    constructor(config, relay) {
        Object.defineProperty(this, "config", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: config
        });
        Object.defineProperty(this, "relay", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: relay
        });
    }
    async login() {
        debug(this.config);
        let loggedIn = false;
        try {
            setTimeout(() => {
                if (!loggedIn)
                    cli_ux_1.default.error(`timed out`);
            }, 1000 * 60 * 10).unref();
            if (process.env.RELAY_API_KEY)
                cli_ux_1.default.error(`Cannot log in with RELAY_API_KEY set`);
            const tokens = session_1.getToken();
            const previousUsername = tokens === null || tokens === void 0 ? void 0 : tokens.username;
            try {
                if (previousUsername) {
                    cli_ux_1.default.warn(`Previously logged in as ${previousUsername}... logging out.`);
                    await this.relay.logout();
                }
            }
            catch (err) {
                cli_ux_1.default.warn(err);
            }
            const auth = await this.browser();
            const subscribers = await this.fetchSubscribers(auth.access_token);
            debug(`subscribers`, subscribers);
            if (!lodash_1.isEmpty(subscribers)) {
                session_1.setToken(auth);
                session_1.saveSubscribers(subscribers);
                const success = await session_1.resolveSubscriber(subscribers);
                if (success) {
                    cli_ux_1.default.log(`Logged in`);
                    loggedIn = true;
                }
                else {
                    cli_ux_1.default.warn(`Default Relay account not set... logging out.`);
                    this.logout();
                }
                return auth;
            }
            else {
                throw new Error(`Failed to discover subscriber id`);
            }
        }
        catch (err) {
            this.logout();
            throw new api_client_1.APIError(err);
        }
    }
    async logout() {
        session_1.deleteSession();
        session_1.clearSubscribers();
    }
    async refresh() {
        const auth = await this.refreshOAuthToken();
        session_1.setToken(auth);
        return auth;
    }
    async browser() {
        // set up callback server
        const { url, codeVerifier } = this.createAuthorization();
        let urlDisplayed = false;
        const showUrl = () => {
            if (!urlDisplayed) {
                cli_ux_1.default.warn(`Cannot open browser.`);
                cli_ux_1.default.warn(`Copy and paste into a browser: ${cli_ux_1.default.url(`Click here to login`, url)}`);
            }
            urlDisplayed = true;
        };
        const { codePromise } = await this.startHttpCodeServer();
        // await cli.open(url)
        debug(url);
        cli_ux_1.default.log(`Opening browser to login`);
        cli_ux_1.default.wait(500);
        const cp = await open_1.default(url, { wait: false });
        cp.on(`error`, err => {
            cli_ux_1.default.warn(err);
            showUrl();
        });
        cp.on(`close`, code => {
            if (code !== 0) {
                showUrl();
            }
        });
        cli_ux_1.default.action.start(`relay: Waiting for login`);
        const code = await codePromise;
        debug(`got code ${code}`);
        cli_ux_1.default.action.stop(`done`);
        const body = {
            grant_type: `authorization_code`,
            redirect_uri: vars_1.vars.authRedirectUri,
            client_id: vars_1.vars.authId,
            code,
            code_verifier: codeVerifier,
        };
        const headers = {
            accept: `application/json`,
            'content-type': `application/x-www-form-urlencoded`,
        };
        const tokenOptions = { headers, body: form_urlencoded_1.default(body) };
        // debug(`token request`, tokenOptions)
        const { body: auth } = await http_call_1.default.post(`${vars_1.vars.authUrl}/oauth2/token`, tokenOptions);
        const validateOptions = { headers: { authorization: `Bearer ${auth.access_token}` } };
        // debug(`validate request`, validateOptions)
        const { body: user } = await http_call_1.default.get(`${vars_1.vars.authUrl}/oauth2/validate`, validateOptions);
        if (!auth.refresh_token) {
            debug(`no refresh_token`);
            const expiresHours = (auth.expires_in || 7200) / 60 / 60;
            cli_ux_1.default.log(`Your login session will expire in ${expiresHours} hours.`);
            // cli.log(`If you sign in again and check "Remember me", your session will not expire`)
            // const prompt = new Toggle({
            //   message: `Sign in again?`,
            //   enabled: `Yes`,
            //   disabled: `No, expire in ${expiresHours} hours`,
            // })
            // if (await prompt.run()) {
            //   return await this.browser()
            // }
        }
        return {
            access_token: auth.access_token,
            refresh_token: auth.refresh_token,
            uuid: user.userid,
            username: user.email,
        };
    }
    startHttpCodeServer() {
        return new Promise((serverResolve, serverReject) => {
            const codePromise = new Promise((codeResolve, codeReject) => {
                const timeout = setTimeout(() => {
                    codeReject(new Error(`timed out waiting for browser login`));
                }, 3 * 60 * 1000); // three minutes
                let socket;
                const server = http_1.default.createServer((req, res) => {
                    clearTimeout(timeout);
                    const respond = (statusCode, message, cb) => {
                        const body = `
              <html>
                <body>
                  <h4>${message}</h4>
                </body>
                <script>
                  window.close()
                </script>
              </html>
            `;
                        res.writeHead(statusCode, {
                            'Content-Length': Buffer.byteLength(body),
                            'Content-Type': `text/html`,
                        });
                        res.flushHeaders();
                        res.write(body);
                        debug(`http responded`);
                        res.end(() => {
                            debug(`http ended`);
                            // @ts-expect-error accessing hidden property `_httpMessage` of socket to expedite server closing
                            const serverResponse = socket._httpMessage;
                            if (serverResponse) {
                                if (!serverResponse.headersSent) {
                                    serverResponse.setHeader(`connection`, `close`);
                                }
                            }
                            socket.destroy();
                            server.close(err => {
                                server.unref();
                                debug(`server closed`);
                                if (err) {
                                    debug(err);
                                }
                                cb();
                            });
                        });
                    };
                    const success = (code) => {
                        respond(200, `Relay CLI authorization complete. You can close this tab.`, () => {
                            codeResolve(code);
                        });
                    };
                    const failure = (status) => {
                        switch (status) {
                            case `no-url`: {
                                respond(418, `Relay CLI authorization failed. Please try again.`, () => {
                                    codeReject(new Error(`redirect contined no url`));
                                });
                                return;
                            }
                            case `not-found`: {
                                respond(404, `Relay CLI authorization failed. Redirected to unknown path. Please try again.`, () => {
                                    codeReject(new Error(`redirected to unknown path`));
                                });
                                return;
                            }
                            case `no-code`: {
                                respond(400, `Relay CLI authorization failed. No code found. Please try again.`, () => {
                                    codeReject(new Error(`authorization code not found on redirect`));
                                });
                                return;
                            }
                        }
                    };
                    if (req === null || req === void 0 ? void 0 : req.url) {
                        const url = new url_1.URL(req.url, vars_1.vars.authRedirectHost);
                        const authUrl = new url_1.URL(vars_1.vars.authRedirectUri);
                        if (url.pathname === authUrl.pathname) {
                            const code = url.searchParams.get(`code`);
                            if (code) {
                                success(code);
                            }
                            else {
                                failure(`no-code`);
                            }
                        }
                        else {
                            failure(`not-found`);
                        }
                    }
                    else {
                        failure(`no-url`);
                    }
                });
                server.on(`connection`, s => {
                    socket = s;
                });
                server.on(`error`, err => {
                    serverReject(err);
                });
                server.listen(vars_1.vars.authRedirectPort, () => {
                    debug(`listening on ${vars_1.vars.authRedirectPort}`);
                    serverResolve({ codePromise });
                });
            });
        });
    }
    createAuthorization() {
        const codeVerifier = utils_1.uuid();
        const codeChallenge = utils_1.base64url(crypto.createHash(`sha256`).update(codeVerifier).digest(`base64`));
        const params = {
            client_id: vars_1.vars.authId,
            response_type: `code`,
            scope: `openid profile`,
            redirect_uri: vars_1.vars.authRedirectUri,
            state: utils_1.uuid(),
            code_challenge_method: `S256`,
            code_challenge: codeChallenge,
        };
        return {
            url: `${vars_1.vars.authUrl}/oauth2/authorization?${querystring_1.default.stringify(params)}`,
            codeVerifier
        };
    }
    async refreshOAuthToken() {
        const tokens = session_1.getToken();
        const refreshToken = tokens === null || tokens === void 0 ? void 0 : tokens.refresh_token;
        if (!refreshToken) {
            throw new Error(`no refresh token to refresh`);
        }
        if (!(tokens === null || tokens === void 0 ? void 0 : tokens.username)) {
            throw new Error(`no username`);
        }
        if (!utils_1.uuid) {
            throw new Error(`no user uuid`);
        }
        const body = {
            grant_type: `refresh_token`,
            refresh_token: refreshToken,
            client_id: vars_1.vars.authId,
        };
        const headers = {
            accept: `application/json`,
            'content-type': `application/x-www-form-urlencoded`,
        };
        const options = { headers, body: form_urlencoded_1.default(body) };
        debug(`refreshOAuthToken`, options);
        const { body: auth } = await http_call_1.default.post(`${vars_1.vars.authUrl}/oauth2/token`, options);
        return {
            ...tokens,
            access_token: auth.access_token,
            refresh_token: auth.refresh_token,
        };
    }
    async fetchSubscribers(token) {
        const options = {
            headers: {
                authorization: `Bearer ${token}`
            }
        };
        const timeout = setTimeout(() => {
            cli_ux_1.default.action.start(`Retrieving authorized subscribers`);
        }, 2000);
        const { body: accounts } = await http_call_1.default.get(`${vars_1.vars.stratusUrl}/v3/subscribers;view=dash_overview`, options);
        clearTimeout(timeout);
        if (cli_ux_1.default.action.running) {
            cli_ux_1.default.action.stop();
        }
        return lodash_1.map(accounts, account => ({
            id: utils_1.getOrThrow(account, [`account`, `subscriber_id`]),
            email: utils_1.getOrThrow(account, [`account`, `owner_email`]),
            name: utils_1.getOrThrow(account, [`account`, `account_name`]),
        }));
    }
}
exports.Login = Login;
