"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.printWorkflows = exports.createTimerWorkflow = exports.createWorkflow = exports.parseArgs = void 0;
const tslib_1 = require("tslib");
const filter_1 = tslib_1.__importDefault(require("lodash/filter"));
const reduce_1 = tslib_1.__importDefault(require("lodash/reduce"));
const flags_1 = require("./flags");
const utils_1 = require("./utils");
const cli_ux_1 = require("cli-ux");
const datetime_1 = require("./datetime");
const parseArgs = (tokens) => {
    /* eslint-disable @typescript-eslint/no-explicit-any */
    const normalArgFlags = filter_1.default(tokens, ({ flag }) => `arg` === flag);
    const normalArgs = reduce_1.default(normalArgFlags, (args, flag) => {
        const [, name, value] = utils_1.parseArg(flag.input);
        return { ...args, [name]: value };
    }, {});
    const booleanParser = flags_1.booleanValue().parse;
    const booleanFlags = filter_1.default(tokens, ({ flag }) => `boolean` === flag);
    const booleanArgs = reduce_1.default(booleanFlags, (args, flag) => {
        const nameValue = booleanParser(flag.input, null);
        return { ...args, ...nameValue };
    }, {});
    const numberParser = flags_1.numberValue().parse;
    const numberFlags = filter_1.default(tokens, ({ flag }) => `number` === flag);
    const numberArgs = reduce_1.default(numberFlags, (args, flag) => {
        const nameValue = numberParser(flag.input, null);
        return { ...args, ...nameValue };
    }, {});
    /* eslint-enable @typescript-eslint/no-explicit-any */
    return { ...normalArgs, ...booleanArgs, ...numberArgs };
};
exports.parseArgs = parseArgs;
const createWorkflow = (flags, tokens) => {
    const args = exports.parseArgs(tokens);
    const workflow = {
        name: flags.name,
        options: {
            transient: flags.transient,
            hidden: flags.hidden,
            allow_remote_invoke: flags.http,
            absorb_triggers: flags.absorb,
        },
        config: {
            trigger: {
                start: {
                    workflow: {
                        uri: flags.uri,
                        args,
                    }
                }
            }
        },
        install: flags.install || []
    };
    return workflow;
};
exports.createWorkflow = createWorkflow;
const createTimerWorkflow = (flags, tokens) => {
    const workflow = exports.createWorkflow(flags, tokens);
    if (flags.trigger === `immediately`) {
        workflow.config.trigger.on_timer = { start_time: `immediately` };
    }
    else if (!flags.start) {
        throw new Error(`Trigger type timer with repeating rule requires specifying a start time. For instance '--start ${datetime_1.getTimestampNow()}'`);
    }
    else {
        const options = {};
        options.start_time = datetime_1.getTimestampFromFlag(flags.start, flags.timezone);
        if (flags.trigger === `schedule`) {
            // schedule... set count to 1
            if (flags.until !== undefined || flags.count !== undefined) {
                throw new Error(`Trigger type schedule runs once in the future and cannot define "count" or "until" values`);
            }
            options.count = 1;
        }
        else {
            // repeat
            if (flags.until && flags.count !== undefined) {
                throw new Error(`Trigger type timer with repeating rule cannot define both a "count" and "until" value`);
            }
            else if (flags.until === undefined && flags.count === undefined) {
                throw new Error(`Trigger type timer with repeating rule must define one of "count" or "until" value`);
            }
            if (flags.until) {
                options.until = datetime_1.getTimestampFromFlag(flags.until, flags.timezone);
            }
            if (flags.frequency) {
                options.frequency = flags.frequency;
            }
            if (flags.interval !== undefined) {
                options.interval = flags.interval;
            }
            if (flags.count !== undefined) {
                options.count = flags.count;
            }
        }
        workflow.config.trigger.on_timer = options;
    }
    return workflow;
};
exports.createTimerWorkflow = createTimerWorkflow;
const printWorkflows = (workflows) => {
    cli_ux_1.cli.styledHeader(`Installed Workflow`);
    cli_ux_1.cli.table(workflows, {
        workflow_id: {
            header: `ID`,
            get: row => row.workflow_id //last(row.workflow_id.split(`_`))
        },
        name: {},
        type: {
            get: utils_1.formatWorkflowType,
        },
        uri: {
            get: row => row.config.trigger.start.workflow.uri,
        },
        args: {
            get: utils_1.formatWorkflowArgs,
        },
        install: {
            header: `Installed on`,
            get: row => row.install.join(`\n`),
        }
    });
};
exports.printWorkflows = printWorkflows;
